=== modified file 'bzr' (properties changed: -x to +x)
=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2005-10-10 21:52:15 +0000
+++ bzrlib/branch.py	2005-10-11 03:27:29 +0000
@@ -358,7 +358,6 @@
             self._lock_count = 1
             self._set_transaction(transactions.PassThroughTransaction())
 
-
     def lock_read(self):
         mutter("lock read: %s (%s)", self, self._lock_count)
         if self._lock_mode:
@@ -371,6 +370,8 @@
             self._lock_mode = 'r'
             self._lock_count = 1
             self._set_transaction(transactions.ReadOnlyTransaction())
+            # 5K may be excessive, but hey, its a knob.
+            self.get_transaction().set_cache_size(5000)
                         
     def unlock(self):
         mutter("unlock: %s (%s)", self, self._lock_count)
@@ -876,8 +877,18 @@
         """Return sequence of revision hashes on to this branch."""
         self.lock_read()
         try:
-            return [l.rstrip('\r\n') for l in
+            transaction = self.get_transaction()
+            history = transaction.map.find_revision_history()
+            if history is not None:
+                mutter("cache hit for revision-history in %s", self)
+                return list(history)
+            history = [l.rstrip('\r\n') for l in
                     self.controlfile('revision-history', 'r').readlines()]
+            transaction.map.add_revision_history(history)
+            # this call is disabled because revision_history is 
+            # not really an object yet, and the transaction is for objects.
+            # transaction.register_clean(history, precious=True)
+            return list(history)
         finally:
             self.unlock()
 
@@ -999,6 +1010,9 @@
         from bzrlib.revision import get_intervening_revisions
         if stop_revision is None:
             stop_revision = other.last_revision()
+        if (stop_revision is not None and 
+            stop_revision in self.revision_history()):
+            return
         greedy_fetch(to_branch=self, from_branch=other,
                      revision=stop_revision)
         pullable_revs = self.missing_revisions(

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2005-10-11 00:36:16 +0000
+++ bzrlib/builtins.py	2005-10-11 04:13:14 +0000
@@ -22,6 +22,7 @@
 import bzrlib.trace
 from bzrlib.trace import mutter, note, log_error, warning
 from bzrlib.errors import BzrError, BzrCheckError, BzrCommandError, NotBranchError
+from bzrlib.errors import DivergedBranches
 from bzrlib.branch import Branch
 from bzrlib import BZRDIR
 from bzrlib.commands import Command
@@ -341,26 +342,25 @@
                 print "Using saved location: %s" % stored_loc
                 location = stored_loc
         cache_root = tempfile.mkdtemp()
-        from bzrlib.errors import DivergedBranches
-        br_from = Branch.open_containing(location)
-        location = br_from.base
-        old_revno = br_to.revno()
+        br_from = Branch.open(location)
+        br_from.lock_read()
         try:
-            from bzrlib.errors import DivergedBranches
-            br_from = Branch.open(location)
             br_from.setup_caching(cache_root)
             location = br_from.base
             old_revno = br_to.revno()
+            old_revision_history = br_to.revision_history()
             try:
                 br_to.update_revisions(br_from)
             except DivergedBranches:
                 raise BzrCommandError("These branches have diverged."
                     "  Try merge.")
-                
-            merge(('.', -1), ('.', old_revno), check_clean=False)
+            new_revision_history = br_to.revision_history()
+            if new_revision_history != old_revision_history:
+                merge(('.', -1), ('.', old_revno), check_clean=False)
             if stored_loc is None or remember:
                 br_to.set_parent(location)
         finally:
+            br_from.unlock()
             rmtree(cache_root)
 
 

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2005-10-10 21:52:15 +0000
+++ bzrlib/fetch.py	2005-10-11 03:19:29 +0000
@@ -117,6 +117,9 @@
             mutter('failed getting last revision: %s', e)
             raise InstallFailed([last_revision])
         mutter('fetch up to rev {%s}', self.last_revision)
+        if (self.last_revision is not None and 
+            self.to_branch.has_revision(self.last_revision)):
+            return
         try:
             revs_to_fetch = self._compare_ancestries()
         except WeaveError:

=== modified file 'bzrlib/identitymap.py'
--- bzrlib/identitymap.py	2005-10-07 04:10:46 +0000
+++ bzrlib/identitymap.py	2005-10-11 02:41:25 +0000
@@ -29,11 +29,25 @@
     that is store in the map. Look for find_CLASS and add_CLASS methods.
     """
 
+    def add_revision_history(self, revision_history):
+        """Add a revision_history object to the map.
+
+        There can only be one!
+        """
+        if self._revision_history is not None:
+            raise errors.BzrError("A revision history (%s) is already "
+                                  "identity map" % self._revision_history)
+        self._revision_history = revision_history
+
     def add_weave(self, id, weave):
         """Add weave to the map with a given id."""
         if self._weave_key(id) in self._map:
             raise errors.BzrError('weave %s already in the identity map' % id)
         self._map[self._weave_key(id)] = weave
+        self._reverse_map[weave] = self._weave_key(id)
+
+    def find_revision_history(self):
+        return self._revision_history
 
     def find_weave(self, id):
         """Return the weave for 'id', or None if it is not present."""
@@ -42,6 +56,20 @@
     def __init__(self):
         super(IdentityMap, self).__init__()
         self._map = {}
+        self._reverse_map = {}
+        self._revision_history = None
+
+    def remove_object(self, an_object):
+        """Remove object from map."""
+        if isinstance(an_object, list):
+            if self._revision_history is an_object:
+                self._revision_history = None
+            else:
+                raise KeyError('%r not in identity map' % an_object)
+        else:
+            self._map.pop(self._reverse_map[an_object])
+            self._reverse_map.pop(an_object)
+        
 
     def _weave_key(self, id):
         """Return the key for a weaves id."""
@@ -57,7 +85,12 @@
     def add_weave(self, id, weave):
         """See IdentityMap.add_weave."""
 
+    def add_revision_history(self, revision_history):
+        """See IdentityMap.add_revision_history."""
+
     def find_weave(self, id):
         """See IdentityMap.find_weave."""
         return None
 
+    def find_revision_history(self):
+        """See IdentityMap.find_revision_history."""

=== modified file 'bzrlib/selftest/testfetch.py'
--- bzrlib/selftest/testfetch.py	2005-10-10 21:52:15 +0000
+++ bzrlib/selftest/testfetch.py	2005-10-11 03:27:29 +0000
@@ -133,7 +133,6 @@
         branch.commit("changed file")
         target = Branch.initialize("target/")
         source = Branch.open(self.get_remote_url("source/"))
-        source.weave_store.enable_cache = False
         self.assertEqual(greedy_fetch(target, source), (2, []))
         # this is the path to the literal file. As format changes 
         # occur it needs to be updated. FIXME: ask the store for the
@@ -145,3 +144,16 @@
         self.assertEqual(1,
             len([log for log in self.weblogs if log.endswith(
                 inventory_weave_suffix)]))
+        # this r-h check test will prevent regressions, but it currently already 
+        # passes, before the patch to cache-rh is applied :[
+        revision_history_suffix = 'revision-history HTTP/1.1" 200 -'
+        self.assertEqual(1,
+            len([log for log in self.weblogs if log.endswith(
+                revision_history_suffix)]))
+        self.weblogs = []
+        # check there is nothing more to fetch
+        source = Branch.open(self.get_remote_url("source/"))
+        self.assertEqual(greedy_fetch(target, source), (0, []))
+        self.failUnless(self.weblogs[0].endswith('branch-format HTTP/1.1" 200 -'))
+        self.failUnless(self.weblogs[1].endswith('revision-history HTTP/1.1" 200 -'))
+        self.assertEqual(2, len(self.weblogs))

=== modified file 'bzrlib/selftest/testidentitymap.py'
--- bzrlib/selftest/testidentitymap.py	2005-10-07 04:10:46 +0000
+++ bzrlib/selftest/testidentitymap.py	2005-10-11 02:41:25 +0000
@@ -46,6 +46,31 @@
         self.assertRaises(errors.BzrError, map.add_weave, "id", weave)
         self.assertEqual(weave, map.find_weave("id"))
  
+    def test_remove_object(self):
+        map = identitymap.IdentityMap()
+        weave = "foo"
+        map.add_weave("id", weave)
+        map.remove_object(weave)
+        map.add_weave("id", weave)
+        rev_history = [1]
+        map.add_revision_history(rev_history)
+        map.remove_object(rev_history)
+
+    def test_add_revision_history(self):
+        map = identitymap.IdentityMap()
+        rev_history = [1,2,3]
+        map.add_revision_history(rev_history)
+        self.assertEqual(rev_history, map.find_revision_history())
+
+    def test_double_add_revision_history(self):
+        map = identitymap.IdentityMap()
+        revision_history = [1]
+        map.add_revision_history(revision_history)
+        self.assertRaises(errors.BzrError,
+                          map.add_revision_history,
+                          revision_history)
+        self.assertEqual(revision_history, map.find_revision_history())
+
  
 class TestNullIdentityMap(TestCase):
 
@@ -68,3 +93,19 @@
         map.add_weave("id", weave)
         self.assertEqual(None, map.find_weave("id"))
         
+    def test_null_identity_map_has_no_remove(self):
+        map = identitymap.NullIdentityMap()
+        self.assertEqual(None, getattr(map, 'remove_object', None))
+
+    def test_add_revision_history(self):
+        map = identitymap.NullIdentityMap()
+        rev_history = [1,2,3]
+        map.add_revision_history(rev_history)
+        self.assertEqual(None, map.find_revision_history())
+
+    def test_double_add_revision_history(self):
+        map = identitymap.NullIdentityMap()
+        revision_history = [1]
+        map.add_revision_history(revision_history)
+        map.add_revision_history(revision_history)
+        self.assertEqual(None, map.find_revision_history())

=== modified file 'bzrlib/selftest/testtransactions.py'
--- bzrlib/selftest/testtransactions.py	2005-10-07 06:13:14 +0000
+++ bzrlib/selftest/testtransactions.py	2005-10-11 02:41:25 +0000
@@ -27,51 +27,130 @@
 import bzrlib.transactions as transactions
 
 
+class DummyWeave(object):
+    """A class that can be instantiated and compared."""
+
+    def __init__(self, message):
+        self._message = message
+
+    def __eq__(self, other):
+        if other is None:
+            return False
+        return self._message == other._message
+
+
+class TestSymbols(TestCase):
+
+    def test_public_symbols(self):
+        from bzrlib.transactions import ReadOnlyTransaction
+        from bzrlib.transactions import PassThroughTransaction
+
+
 class TestReadOnlyTransaction(TestCase):
 
-    def test_symbols(self):
-        from bzrlib.transactions import ReadOnlyTransaction
-
-    def test_construct(self):
-        transactions.ReadOnlyTransaction()
+    def setUp(self):
+        self.transaction = transactions.ReadOnlyTransaction()
+        super(TestReadOnlyTransaction, self).setUp()
 
     def test_register_clean(self):
-        transaction = transactions.ReadOnlyTransaction()
-        transaction.register_clean("anobject")
+        self.transaction.register_clean("anobject")
 
     def test_register_dirty_raises(self):
-        transaction = transactions.ReadOnlyTransaction()
         self.assertRaises(errors.ReadOnlyError, 
-                          transaction.register_dirty,"anobject")
+                          self.transaction.register_dirty,"anobject")
     
     def test_commit_raises(self):
-        transaction = transactions.ReadOnlyTransaction()
-        self.assertRaises(errors.CommitNotPossible, transaction.commit)
+        self.assertRaises(errors.CommitNotPossible, self.transaction.commit)
 
     def test_map(self):
-        transaction = transactions.ReadOnlyTransaction()
-        self.assertNotEqual(None, getattr(transaction, "map", None))
+        self.assertNotEqual(None, getattr(self.transaction, "map", None))
     
     def test_add_and_get(self):
-        transaction = transactions.ReadOnlyTransaction()
         weave = "a weave"
-        transaction.map.add_weave("id", weave)
-        self.assertEqual(weave, transaction.map.find_weave("id"))
+        self.transaction.map.add_weave("id", weave)
+        self.assertEqual(weave, self.transaction.map.find_weave("id"))
 
     def test_rollback_returns(self):
-        transaction = transactions.ReadOnlyTransaction()
-        transaction.rollback()
+        self.transaction.rollback()
 
     def test_finish_returns(self):
-        transaction = transactions.ReadOnlyTransaction()
-        transaction.finish()
+        self.transaction.finish()
 
+    def test_zero_size_cache(self):
+        self.transaction.set_cache_size(0)
+        weave = DummyWeave('a weave')
+        self.transaction.map.add_weave("id", weave)
+        self.assertEqual(weave, self.transaction.map.find_weave("id"))
+        weave = None
+        # add an object, should fall right out if there are no references
+        self.transaction.register_clean(self.transaction.map.find_weave("id"))
+        self.assertEqual(None, self.transaction.map.find_weave("id"))
+        # but if we have a reference it should stick around
+        weave = DummyWeave("another weave")
+        self.transaction.map.add_weave("id", weave)
+        self.transaction.register_clean(self.transaction.map.find_weave("id"))
+        self.assertEqual(weave, self.transaction.map.find_weave("id"))
+        del weave
+        # its not a weakref system
+        self.assertEqual(DummyWeave("another weave"),
+                         self.transaction.map.find_weave("id"))
         
+    def test_small_cache(self):
+        self.transaction.set_cache_size(1)
+        # add an object, should not fall right out if there are no references
+        #sys.getrefcounts(foo)
+        self.transaction.map.add_weave("id", DummyWeave("a weave"))
+        self.transaction.register_clean(self.transaction.map.find_weave("id"))
+        self.assertEqual(DummyWeave("a weave"),
+                         self.transaction.map.find_weave("id"))
+        self.transaction.map.add_weave("id2", DummyWeave("a weave also"))
+        self.transaction.register_clean(self.transaction.map.find_weave("id2"))
+        # currently a fifo
+        self.assertEqual(None, self.transaction.map.find_weave("id"))
+        self.assertEqual(DummyWeave("a weave also"),
+                         self.transaction.map.find_weave("id2"))
+
+    def test_small_cache_with_references(self):
+        # if we have a reference it should stick around
+        weave = "a weave"
+        weave2 = "another weave"
+        self.transaction.map.add_weave("id", weave)
+        self.transaction.map.add_weave("id2", weave2)
+        self.assertEqual(weave, self.transaction.map.find_weave("id"))
+        self.assertEqual(weave2, self.transaction.map.find_weave("id2"))
+        weave = None
+        # its not a weakref system
+        self.assertEqual("a weave", self.transaction.map.find_weave("id"))
+
+    def test_precious_with_zero_size_cache(self):
+        self.transaction.set_cache_size(0)
+        weave = DummyWeave('a weave')
+        self.transaction.map.add_weave("id", weave)
+        self.assertEqual(weave, self.transaction.map.find_weave("id"))
+        weave = None
+        # add an object, should not fall out even with no references.
+        self.transaction.register_clean(self.transaction.map.find_weave("id"),
+                                        precious=True)
+        self.assertEqual(DummyWeave('a weave'),
+                         self.transaction.map.find_weave("id"))
+
+    def test_precious_revision_history(self):
+        """Disabled test until revision-history is a real object."""
+        print "Disabled: test_precious_revision_history"
+        return
+        self.transaction.set_cache_size(0)
+        history = []
+        self.transaction.map.add_revision_history(history)
+        self.assertEqual(history, self.transaction.map.find_revision_history())
+        history = None
+        # add an object, should not fall out even with no references.
+        self.transaction.register_clean(
+            self.transaction.map.find_revision_history(), precious=True)
+        self.assertEqual([], self.transaction.map.find_revision_history())
+
+
 class TestPassThroughTransaction(TestCase):
 
-    def test_symbols(self):
-        from bzrlib.transactions import PassThroughTransaction
-
     def test_construct(self):
         transactions.PassThroughTransaction()
 
@@ -104,3 +183,10 @@
     def test_finish_returns(self):
         transaction = transactions.PassThroughTransaction()
         transaction.finish()
+
+    def test_cache_is_ignored(self):
+        transaction = transactions.PassThroughTransaction()
+        transaction.set_cache_size(100)
+        weave = "a weave"
+        transaction.map.add_weave("id", weave)
+        self.assertEqual(None, transaction.map.find_weave("id"))

=== modified file 'bzrlib/store/weave.py'
--- bzrlib/store/weave.py	2005-10-10 03:24:15 +0000
+++ bzrlib/store/weave.py	2005-10-11 01:05:24 +0000
@@ -39,9 +39,10 @@
     """
     FILE_SUFFIX = '.weave'
 
-    def __init__(self, transport, prefixed=False):
+    def __init__(self, transport, prefixed=False, precious=False):
         self._transport = transport
         self._prefixed = prefixed
+        self._precious = precious
 
     def filename(self, file_id):
         """Return the path relative to the transport root."""
@@ -78,7 +79,7 @@
             return weave
         w = read_weave(self._get(file_id))
         transaction.map.add_weave(file_id, w)
-        transaction.register_clean(w)
+        transaction.register_clean(w, precious=self._precious)
         return w
 
     def get_lines(self, file_id, rev_id, transaction):
@@ -95,7 +96,7 @@
         except NoSuchFile:
             weave = Weave(weave_name=file_id)
             transaction.map.add_weave(file_id, weave)
-            transaction.register_clean(weave)
+            transaction.register_clean(weave, precious=self._precious)
             return weave
 
     def put_weave(self, file_id, weave, transaction):

=== modified file 'bzrlib/transactions.py'
--- bzrlib/transactions.py	2005-10-07 06:13:14 +0000
+++ bzrlib/transactions.py	2005-10-11 01:05:24 +0000
@@ -39,8 +39,11 @@
 repeated reads cannot be obtained if the object is ejected.
 """
 
+import sys
+
 import bzrlib.errors as errors
 from bzrlib.identitymap import IdentityMap, NullIdentityMap
+from bzrlib.trace import mutter
 
 
 class ReadOnlyTransaction(object):
@@ -58,10 +61,24 @@
         """
 
     def __init__(self):
+        super(ReadOnlyTransaction, self).__init__()
         self.map = IdentityMap()
+        self._clean_objects = set()
+        self._clean_queue = []
+        self._limit = -1
+        self._precious_objects = set()
 
-    def register_clean(self, an_object):
-        """Register an_object as being clean."""
+    def register_clean(self, an_object, precious=False):
+        """Register an_object as being clean.
+        
+        If the precious hint is True, the object will not
+        be ejected from the object identity map ever.
+        """
+        self._clean_objects.add(an_object)
+        self._clean_queue.append(an_object)
+        if precious:
+            self._precious_objects.add(an_object)
+        self._trim()
 
     def register_dirty(self, an_object):
         """Register an_object as being dirty."""
@@ -71,6 +88,39 @@
     def rollback(self):
         """Let people call this even though nothing has to happen."""
 
+    def set_cache_size(self, size):
+        """Set a new cache size."""
+        assert -1 <= size
+        self._limit = size
+        self._trim()
+
+    def _trim(self):
+        """Trim the cache back if needed."""
+        if self._limit < 0 or self._limit - len(self._clean_objects) > 0:
+            return
+        needed = len(self._clean_objects) - self._limit
+        offset = 0
+        while needed and offset < len(self._clean_objects):
+            # references we know of:
+            # temp passed to getrefcount in our frame
+            # temp in getrefcount's frame
+            # the map forward
+            # the map backwards
+            # _clean_objects
+            # _clean_queue
+            # 1 missing ?
+            if (sys.getrefcount(self._clean_queue[offset]) <= 7 and
+                not self._clean_queue[offset] in self._precious_objects):
+                removed = self._clean_queue[offset]
+                self._clean_objects.remove(removed)
+                del self._clean_queue[offset]
+                self.map.remove_object(removed)
+                mutter('removed object %r', removed)
+                needed -= 1
+            else:
+                offset += 1
+
+
         
 class PassThroughTransaction(object):
     """A pass through transaction
@@ -91,10 +141,15 @@
         """
 
     def __init__(self):
+        super(PassThroughTransaction, self).__init__()
         self.map = NullIdentityMap()
 
-    def register_clean(self, an_object):
-        """Register an_object as being clean."""
+    def register_clean(self, an_object, precious=False):
+        """Register an_object as being clean.
+        
+        Note that precious is only a hint, and PassThroughTransaction
+        ignores it.
+        """
 
     def register_dirty(self, an_object):
         """Register an_object as being dirty."""
@@ -102,3 +157,6 @@
     def rollback(self):
         """Cannot rollback a pass through transaction."""
         raise errors.AlreadyCommitted
+
+    def set_cache_size(self, ignored):
+        """Do nothing, we are passing through."""

